/*
 * Copyright 2005-2010 NVIDIA Corporation.  All rights reserved.
 */

// ====================================================================
//
// Module: opt_live.cxx
//
//    This module is for implementing Live Analysis on
//    SSA-based IR. The whole analysis is implemented as a
//    class SSA_Live. The result of the analysis is
//    stored in the class, and several interfaces are provided to
//    retrieve the information for each BB_NODE, or the whole procedure. 
// ====================================================================
//
#include "opt_live.h"
#include "opt_defs.h"
#include <list>
#include "tracing.h"

#define DEP_CHECKED 1
#define ON_STACK 2

SSA_Live::SSA_Live(COMP_UNIT *cunit, SSA_Refcount *cr_ref)
{
  _comp_unit = cunit;
  MEM_POOL_Initialize(&_mem_pool, "SSA_Live_Pool", FALSE);
  MEM_POOL_Push(&_mem_pool);
  _topVarNumber = cr_ref->Generate_Var_Map(_Cr2LiveMap, _regCrMap);
  Build_live_in();
}

SSA_Live::~SSA_Live(void)
{
  MEM_POOL_Pop(&_mem_pool);
  MEM_POOL_Delete(&_mem_pool);
}

// Collect uses in an expression recursively
void
SSA_Live::Collect_use_rec(CODEREP *cr, BOOL is_store, BB_LIVE_INFO *live_info)
{
  if (cr->Kind() == CK_VAR) {
    Mark_up_exposed(cr, live_info); 
  }
  else if(cr->Kind() == CK_IVAR){

    MU_NODE *mnode = cr->Ivar_mu_node();
    if (mnode != NULL)
      Mark_up_exposed(mnode->OPND(), live_info);

    if (cr->Opr() == OPR_PARM && (cr->Offset() & WN_PARM_DUMMY))
      return;

    Collect_use_rec(is_store ? cr->Istr_base() : cr->Ilod_base(), FALSE, live_info);

    if (cr->Opr() == OPR_MLOAD) {
      Collect_use_rec(cr->Mload_size() ?  cr->Mload_size() : cr->Mstore_size(),
                        FALSE, live_info);
    }
    else if (cr->Opr() == OPR_ILOADX) {
      Collect_use_rec(cr->Index(), FALSE, live_info);
    }
  } else if (cr->Kind() == CK_OP) {
    for (INT i=0; i< cr->Kid_count(); i++) {
      Collect_use_rec(cr->Opnd(i), FALSE, live_info);
    }
  }
}

// Collect initial live information in a BB_NODE
void SSA_Live::init_bb_live_info(BB_NODE *bb)
{
  BB_LIVE_INFO *live_info = _BBLiveMap[bb];

  STMTREP_ITER stmt_iter(bb->Stmtlist());
  STMTREP *stmt;
  FOR_ALL_NODE_REVERSE(stmt, stmt_iter, Init()) {

    OPERATOR opr= stmt->Opr();
    if (opr==OPR_STID) {
      Mark_killed(stmt->Lhs(), live_info);
    }

    if( stmt->Has_chi()) {
      CHI_LIST *chi_list = stmt->Chi_list();
      CHI_NODE *cnode;
      CHI_LIST_ITER chi_iter;
      FOR_ALL_NODE(cnode, chi_iter, Init(chi_list)) {
        if (cnode->Live()) {
          Mark_killed(cnode->RESULT(), live_info);
          Mark_up_exposed(cnode->OPND(), live_info);
        }
      }
    }

    if(stmt->Has_mu()) {
      MU_LIST *mu_list= stmt->Mu_list();
      if (mu_list!=NULL)  {
        MU_NODE *mnode;
        MU_LIST_ITER mu_iter;
        FOR_ALL_NODE(mnode, mu_iter, Init(mu_list))  {
          Mark_up_exposed(mnode->OPND(), live_info);
        }
      }
    }
 

    if (opr != OPR_STID && stmt->Lhs() != NULL) {
      Collect_use_rec(stmt->Lhs(), TRUE, live_info);
    }

    if (stmt->Rhs())
      Collect_use_rec(stmt->Rhs(), FALSE, live_info);
  }

  live_info->live_in->CopyD(live_info->live_gen);
}

// Mark a CODEREP killed in a BB_NODE
void SSA_Live::Mark_killed(CODEREP *kill_cr, BB_LIVE_INFO *live_info)
{
  INT32 kill_id = getCRLiveNumber(kill_cr);
  if (kill_id < 0) return;

  live_info->live_kill->Union1D(kill_id);

  live_info->live_gen->Difference1D(kill_id);
}

// Mark a CODEREP up-exposed in a BB_NODE
void SSA_Live::Mark_up_exposed(CODEREP *use_cr, BB_LIVE_INFO *live_info)
{
  INT32 use_id = getCRLiveNumber(use_cr);
  if (use_id < 0) return;

  live_info->live_gen->Union1D(use_id);
}

// Build live information for the current comp_unit
void SSA_Live::Build_live_in(void)
{
  // step 1: initialize Live map: each CK_VAR CR map to a unique Live number
  CODEREP *cr;
  CODEREP_ITER cr_iter;
  AUX_ID i;
  
  _liveSetSize = _topVarNumber;

  CFG_ITER *c_iter = new CFG_ITER(_comp_unit->Cfg());

  for (BB_NODE *bb = c_iter->First(); bb; bb = c_iter->Next()) {
    _BBLiveMap[bb] = new BB_LIVE_INFO(_liveSetSize, &_mem_pool);
  }

  for (BB_NODE *bb = c_iter->First(); bb; bb = c_iter->Next()) {
    PHI_NODE *phi;
    PHI_LIST_ITER phi_iter;
    FOR_ALL_ELEM (phi, phi_iter, Init(bb->Phi_list())) {
      if (phi->Live()) {
        CODEREP *re = phi->RESULT();
        Mark_up_exposed(re, _BBLiveMap[bb]);
        BB_LIST_ITER bb_iter;
        BB_NODE *bb_phi;
        int pos = 0;
        FOR_ALL_ELEM (bb_phi, bb_iter, Init(bb->Pred())) 
        {
           Mark_killed(re, _BBLiveMap[bb]);
           Mark_up_exposed(phi->OPND(pos), _BBLiveMap[bb_phi]);
           pos++;
        }
      }
    }
  }

  for (BB_NODE *bb = c_iter->First(); bb; bb = c_iter->Next()) {
    init_bb_live_info(bb);
  }

  bool changed = true;

  IDX_32_SET *new_live_out = new IDX_32_SET(_topVarNumber, &_mem_pool, OPTS_FALSE);

  _comp_unit->Cfg()->Pdo_vec();
  INT32 size = _comp_unit->Cfg()->Pdo_vec_sz();

  while (changed) {
    changed = false;
    // traverse the CFG with Post-dominator order
    for (INT32 i = 0; i < size; i++) {
      BB_NODE *bb = _comp_unit->Cfg()->Pdo_Bb(i);
      BB_NODE *succ;
      BB_LIST_ITER bb_iter;
      bool first = true;
      BB_LIVE_INFO *live_info = _BBLiveMap[bb];
      FOR_ALL_ELEM (succ, bb_iter, Init(bb->Succ())) {
        BB_LIVE_INFO *succ_live_info = _BBLiveMap[succ];
        if (first) {
          first = false;
          new_live_out->CopyD(succ_live_info->live_in);
        } else {
          new_live_out->UnionD(succ_live_info->live_in);
        } 
      }

      if (!new_live_out->EqualP(live_info->live_out)) {
        changed = true;
        IDX_32_SET *tmp = live_info->live_out;
        live_info->live_out = new_live_out;
        new_live_out = tmp;
        live_info->live_in->CopyD(live_info->live_out);
        live_info->live_in->DifferenceD(live_info->live_kill)->UnionD(live_info->live_gen);
      }
    } 
  }


  Compute_max_live_in ();
}

// Get the Live number for a given CODEREP
INT32 SSA_Live::getCRLiveNumber(CODEREP *cr) 
{
  if (_Cr2LiveMap.find(cr) != _Cr2LiveMap.end())
    return _Cr2LiveMap[cr];

  return -1;
}

INT32 SSA_Live::Compute_bb_live_in(BB_NODE *bb)
{
  INT32 count;
  BB_LIVE_INFO *blive = _BBLiveMap[bb];

  count = 0;
  for (IDX_32 x = blive->live_in->Choose(); x != IDX_32_SET_CHOOSE_FAILURE;
       x = blive->live_in->Choose_Next(x)) {
    count++;
  }
  blive->_live_in_count = count;
  return count;
}

BOOL SSA_Live::CR_is_live_out(CODEREP *cr, BB_LIVE_INFO *blive) 
{
  if (cr->Kind() != CK_VAR)
    return FALSE;

  INT32 idx = getCRLiveNumber(cr);
  if (idx < 0) 
    return FALSE;

  return blive->live_out->MemberP(idx);
}

BOOL SSA_Live::CR_is_live_in(CODEREP *cr, BB_LIVE_INFO *blive) 
{
  if (cr->Kind() != CK_VAR)
    return FALSE;

  INT32 idx = getCRLiveNumber(cr);
  if (idx < 0)
    return FALSE;

  return blive->live_in->MemberP(idx);
}

INT32 SSA_Live::Compute_bb_max_live_count(BB_NODE *bb)
{
  INT32 live_in_count;
  BB_LIVE_INFO *blive = _BBLiveMap[bb];

  live_in_count = 0;
  std::set<AUX_ID> auxSet;
  INT32 count = 0;
  for (IDX_32 x = blive->live_in->Choose(); x != IDX_32_SET_CHOOSE_FAILURE;
       x = blive->live_in->Choose_Next(x)) {
    CODEREP *cr = _regCrMap[x];
    if (cr == NULL) continue;
    if (auxSet.find(cr->Aux_id()) != auxSet.end())
      continue;
    auxSet.insert(cr->Aux_id());
    live_in_count += MTYPE_byte_size(cr->Dtyp())/4;
    count++;
  }

  if (live_in_count > count) {
    // discount the double or complex effect on live-count
    live_in_count = count + (2 * (live_in_count - count) )/3;
  }

  return live_in_count;
}

// Compute the number of Live-in for a BB_NODE
INT32 SSA_Live::Get_bb_live_in(BB_NODE *bb) 
{
  BB_LIVE_INFO *blive = _BBLiveMap[bb];

  return blive->_live_in_count;
}

// Compute the Max-live-in of the current comp_unit
INT32 SSA_Live::Get_max_live_in(void) const 
{
  return _max_live_in_count;
}

INT32 SSA_Live::Compute_max_live_in()
{
  INT32 maxResult = 0;
  CFG_ITER *c_iter = new CFG_ITER(_comp_unit->Cfg());
  for (BB_NODE *bb = c_iter->First(); bb; bb = c_iter->Next()) {
    INT32 count = Compute_bb_max_live_count(bb);
    Is_Trace(Get_Trace(TP_WOPT3, EPRE_REG_PRESSURE_FLAG), (TFile, "B[%d] = %d\n", bb->Id(), count));

    if (maxResult < count) maxResult = count;
  }

  _max_live_in_count = maxResult;

  return maxResult;
}

// External interface for just computing the Max live-in
INT32 Compute_SSA_Max_Live(COMP_UNIT *comp_unit, SSA_Refcount *ref)
{
  INT32 count;

  SSA_Live *ssa_live = new SSA_Live(comp_unit, ref);

  count = ssa_live->Get_max_live_in();
  delete ssa_live;

  return count;
}

// Track variables and expressions reference count in a procedure
SSA_Refcount::SSA_Refcount(COMP_UNIT *comp_unit)
{
  CFG_ITER *c_iter = new CFG_ITER(comp_unit->Cfg());
  _var_count = 0;

  for (BB_NODE *bb = c_iter->First(); bb; bb = c_iter->Next()) {
    BB_init_cr_refcount(bb);
  }
}

// Collect reference counts in an expression    
void
SSA_Refcount::CR_collect_refcount(CODEREP *cr)
{
  if (cr->Kind() == CK_VAR) {
    CR_Inc_refcount(cr);
  }
  else if(cr->Kind() == CK_IVAR){
    CR_Inc_refcount(cr);
    if (cr->Ilod_base()) {
      CR_collect_refcount(cr->Ilod_base());
    } else if (cr->Istr_base()) {
      CR_collect_refcount(cr->Istr_base());
    }

    if (cr->Opr() == OPR_MLOAD) {
      if (cr->Mload_size()) {
        CR_collect_refcount(cr->Mload_size());
      } else if (cr->Mstore_size()) {
        CR_collect_refcount(cr->Mstore_size());
      }
    }
    else if (cr->Opr() == OPR_ILOADX) {
      if (cr->Index()) {
        CR_collect_refcount(cr->Index());
      }
    }
  } else if (cr->Kind() == CK_OP) {
    CR_Inc_refcount(cr);
    for (INT i=0; i< cr->Kid_count(); i++) {
      CR_collect_refcount(cr->Opnd(i));
    }
  }
}

// Track reference count in a statement
void SSA_Refcount::Stmt_Inc_refcount(STMTREP *stmt)
{
  OPERATOR opr= stmt->Opr();
  if (opr != OPR_STID && stmt->Lhs() != NULL) {
    CR_collect_refcount(stmt->Lhs());
  }

  if (stmt->Rhs())
    CR_collect_refcount(stmt->Rhs());
}

// track reference counts and mark CRs which could not be in registers
void SSA_Refcount::Stmt_init_refcount(STMTREP *stmt)
{
  OPERATOR opr= stmt->Opr();
  if (opr != OPR_STID && stmt->Lhs() != NULL) {
    CR_collect_refcount(stmt->Lhs());
  }

  if (stmt->Rhs())
    CR_collect_refcount(stmt->Rhs());

  if (opr == OPR_STID && stmt->Chi_list() != NULL) {
    CR_mark_non_reg(stmt->Lhs());
  }

  if (opr == OPR_EVAL && stmt->Rhs() && stmt->Rhs()->Kind() == CK_VAR) {
    CR_mark_non_reg(stmt->Rhs());
  }
}

// track and decrease reference counts in a statement
void SSA_Refcount::Stmt_Dec_refcount(STMTREP *stmt)
{
  OPERATOR opr= stmt->Opr();
  if (opr != OPR_STID && stmt->Lhs() != NULL) {
    Expr_Dec_refcount(stmt->Lhs());
  }

  if (stmt->Rhs())
    Expr_Dec_refcount(stmt->Rhs());
}

// Track reference counts in a BB_NODE
void SSA_Refcount::BB_init_cr_refcount(BB_NODE *bb)
{
  STMTREP_ITER stmt_iter(bb->Stmtlist());
  STMTREP *stmt;

  FOR_ALL_NODE_REVERSE(stmt, stmt_iter, Init()) {
    Stmt_init_refcount(stmt);
  }
}

